//
// Created by kc on 1/03/21.
//

#ifndef BSETSALT_ARCHIVE_H
#define BSETSALT_ARCHIVE_H

#include <vector>
#include "PointSet.h"


/**
 * Represents a set of non-equivalent point sets of the same size. Used during generation to
 * check whether a newly generated set is equivalent to a set that was generated earlier.
 *
 * Uses invariants to avoid unnecessary computation of the canonical form of its members.
 *
 * We keep track of one archive for each set size. This must be cleared explicitly after use.
 *
 */
class Archive {

private:
    // private class, represents a member of the archive - do not use separately
    class Member {
    public:
        PointSet set;

        int invariant;

        SparseGraph *canonicalForm; // lazily computed

        Member(PointSet set, int invariant, SparseGraph *canonicalForm)
                : set(set), invariant(invariant), canonicalForm(canonicalForm) {};

        // Clear memory used by canonical form
        void clear() const {
            delete canonicalForm; // note: deleting nullptr has no effect
        }

        // Return the canonical form, lazily computed
        const SparseGraph *getCanonicalForm();

        // Does this member represent a point set equivalent to the given point set?
        bool isSameAs(Member &other);

    };

    std::vector<Member> members;

public:

    // Add the given point set to the archive, with the given invariant value and canonical form
    // If an equivalent point set is already in the archive, then no action is taken and false is returned, otherwise
    // the point set is added, and true is returned
    bool add(PointSet &set, int invariant, SparseGraph *canonicalForm);

    // Same as above but with unknown canonical form

    bool add(PointSet &set, int invariant) {
        return add(set, invariant, nullptr);
    }

    // clears all members of the archive when the archive is no longer needed
    ~Archive() {
        for (const Member &member : members) {
            member.clear();
        }
    }

};

#endif //BSETSALT_ARCHIVE_H
