#include <iostream>
#include "Generator.h"

#include "Archive.h"
#include "ColouredPlane.h"

void Generator::generate() {

    // start with a two point subset. All two point subsets are equivalent/isomorphic
    add(0);
    add(1);

    gen();

    remove(1);
    remove(0); // these were added to be able to check memory leaks

}

// recursive version of generate
void Generator::gen() {
    // bail out when not in the correct search branch (when parallelized)
    if (currentSet.getSize() == splitDepth) {
        if (splitCounter++ % splitModulus != splitIndex) {
            return;
        }
    }

    if (isComplete()) {
//        currentSet.print();
        totalCount ++;
        counts[currentSet.getSize()] ++;
    } else {
        Archive archive;
        for (int pt = 0; pt < V; pt++) {
            if (!currentSet.contains(pt)) {
                if (isValid(pt)) {
                    add(pt);
                    if (isValid()) {
                        switch (checkMaybeCanonical(pt)) {
                            case CANONICAL: {
                                if (archive.add(currentSet, weights.computeSetInvariant())) {
                                    gen();
                                }
                                break;
                            }
                            case UNKNOWN: {
                                SparseGraph *canonicalForm = ColouredPlane(currentSet).isNautyCanonical(pt, weights);
                                if (canonicalForm) { // yes, it is special!
                                    if (archive.add(currentSet, weights.computeSetInvariant(),
                                                    canonicalForm)) {
                                        gen();
                                    }
                                }
                                break;
                            }
                            case NOT_CANONICAL: {
                                // do not recurse
                                break;
                            }
                        }
                    }
                    remove(pt);
                }
            }
        }
    }
}

// check whether the currently generated set is complete
bool Generator::isComplete() {
    return weights.isBlocking(); // for testing
}

// can this point be added to the set so that it remains a valid parent?
bool Generator::isValid(int pt) {
    return ! weights.isSuperfluous(pt);
}

// is the given point canonical in the set (wrt. isomorph freeness)
CheckResult Generator::checkMaybeCanonical(int pt) {
    weights.computeInvariants();
    return weights.checkCanonical(pt);
}

void Generator::add(int pt) {
    currentSet.add(pt);
    weights.add(pt);
}

void Generator::remove(int pt) {
    weights.remove(pt);
    currentSet.remove(pt);
}

bool Generator::isValid() {
    return weights.isMinimal();
}

void Generator::printStatistics() {
    std::cerr << "Blocking sets of PG(2," << Q << ")" << std::endl;
    if (splitDepth <= V) {
        std::cerr << " parallellised - level " << splitDepth << ", index " << splitIndex << ", modulus " << splitModulus << std::endl;
    }
    int start = 0;
    while (counts[start] == 0) {
        start ++;
    }
    int stop = V-1;
    while (counts[stop] == 0) {
        stop --;
    }

    for (int i=start; i <= stop; i++) {
        std::cerr << "Size " << i << ": " << counts[i] << std::endl;
    }
}

void Generator::processArgs(int argc, char **argv) {
    if (argc > 1) {
        if (argc != 4) {
            std::cerr << "Usage: " << argv[0] << " depth index modulus";
        } else {
            splitDepth = (int) strtol(argv[1], nullptr, 10);
            splitIndex = (int) strtol(argv[2], nullptr, 10);
            splitModulus = (int) strtol(argv[3], nullptr, 10);
        }
    }

}



